"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __classPrivateFieldSet = (this && this.__classPrivateFieldSet) || function (receiver, privateMap, value) {
    if (!privateMap.has(receiver)) {
        throw new TypeError("attempted to set private field on non-instance");
    }
    privateMap.set(receiver, value);
    return value;
};
var __classPrivateFieldGet = (this && this.__classPrivateFieldGet) || function (receiver, privateMap) {
    if (!privateMap.has(receiver)) {
        throw new TypeError("attempted to get private field on non-instance");
    }
    return privateMap.get(receiver);
};
var _priority, _inner, _pendingInsertion, _subscriptions, _resolvers;
Object.defineProperty(exports, "__esModule", { value: true });
exports.PendingInsertionError = exports.ResolvePriority = void 0;
var ResolvePriority;
(function (ResolvePriority) {
    ResolvePriority["Takes"] = "takes";
    ResolvePriority["Loads"] = "loads";
})(ResolvePriority = exports.ResolvePriority || (exports.ResolvePriority = {}));
class PendingInsertionError extends Error {
    constructor() {
        super("cell is currently waiting for pending insertions");
    }
}
exports.PendingInsertionError = PendingInsertionError;
/**
 * A cell that holds a value and can alert users when a value in inserted.
 */
class AsyncCell {
    constructor(value, priority = ResolvePriority.Takes) {
        _priority.set(this, void 0);
        _inner.set(this, void 0);
        _pendingInsertion.set(this, false);
        _subscriptions.set(this, []);
        _resolvers.set(this, {
            takes: [],
            loads: [],
        });
        __classPrivateFieldSet(this, _inner, value);
        __classPrivateFieldSet(this, _priority, priority);
    }
    /**
     * Returns the value from the cell or waits for a value to be inserted.
     * @returns the value in the cell.
     */
    load() {
        if (__classPrivateFieldGet(this, _inner) !== undefined) {
            return Promise.resolve(__classPrivateFieldGet(this, _inner));
        }
        return new Promise((resolve) => {
            __classPrivateFieldGet(this, _resolvers).loads.push(resolve);
        });
    }
    /**
     * Waits for a value to be placed in the cell and then resets it's inner value.
     * @returns the value placed into the cell.
     */
    take() {
        return __awaiter(this, void 0, void 0, function* () {
            if (__classPrivateFieldGet(this, _inner) !== undefined) {
                const oldValue = __classPrivateFieldGet(this, _inner);
                __classPrivateFieldSet(this, _inner, undefined);
                return oldValue;
            }
            return yield new Promise((resolve) => {
                __classPrivateFieldGet(this, _resolvers).takes.push(resolve);
            });
        });
    }
    /**
     * Peeks into the cell to check the current inner value.
     * @returns the current inner value of the cell.
     */
    peek() {
        return __classPrivateFieldGet(this, _inner);
    }
    /**
     * Inserts a value into the cell and resolves all promises waiting on the cell to be filled.
     * @param value the value to insert into the cell.
     */
    insert(value) {
        if (__classPrivateFieldGet(this, _pendingInsertion)) {
            throw new PendingInsertionError();
        }
        const prevValue = __classPrivateFieldGet(this, _inner);
        if (value === undefined) {
            __classPrivateFieldSet(this, _inner, undefined);
            __classPrivateFieldGet(this, _subscriptions).forEach((subscription) => subscription(undefined, prevValue));
        }
        else {
            __classPrivateFieldSet(this, _inner, value);
            __classPrivateFieldGet(this, _subscriptions).forEach((subscription) => subscription(value, prevValue));
        }
        if (__classPrivateFieldGet(this, _inner) === undefined)
            return;
        // If we prioritize loads we'll resolve them before we resolve takes.
        if (__classPrivateFieldGet(this, _priority) === ResolvePriority.Loads) {
            __classPrivateFieldGet(this, _resolvers).loads.forEach((resolver) => resolver(__classPrivateFieldGet(this, _inner)));
            __classPrivateFieldGet(this, _resolvers).loads = [];
        }
        const takeResolveFunction = __classPrivateFieldGet(this, _resolvers).takes.shift();
        if (takeResolveFunction !== undefined) {
            takeResolveFunction(__classPrivateFieldGet(this, _inner));
        }
        else if (__classPrivateFieldGet(this, _priority) === ResolvePriority.Takes) {
            // Otherwise if we prioritize takes we'll resolve them after we do any takes, but not if there were tany takes.
            __classPrivateFieldGet(this, _resolvers).loads.forEach((resolver) => resolver(__classPrivateFieldGet(this, _inner)));
            __classPrivateFieldGet(this, _resolvers).loads = [];
        }
    }
    /**
     * Runs the provided function if no value is currently in the cell. If the result of the function
     * is a promise, the cell will block any insertions until it has resolved.
     * @param func a function returning a value to insert in the cell.
     */
    orInsert(func) {
        if (__classPrivateFieldGet(this, _inner) !== undefined)
            return;
        if (__classPrivateFieldGet(this, _pendingInsertion))
            throw new PendingInsertionError();
        const output = func();
        if (output instanceof Promise) {
            __classPrivateFieldSet(this, _pendingInsertion, true);
            output.then((value) => {
                __classPrivateFieldSet(this, _pendingInsertion, false);
                this.insert(value);
            });
        }
        else {
            this.insert(output);
        }
    }
    /**
     * Registers a {@link Subscription} that is called when a new value is inserted into the cell.
     * @param subscriptions subscriptions to be invoked when a new value is inserted into the cell.
     */
    subscribe(...subscriptions) {
        __classPrivateFieldSet(this, _subscriptions, [...__classPrivateFieldGet(this, _subscriptions), ...subscriptions]);
    }
}
exports.default = AsyncCell;
_priority = new WeakMap(), _inner = new WeakMap(), _pendingInsertion = new WeakMap(), _subscriptions = new WeakMap(), _resolvers = new WeakMap();
